#ifndef SBUF_I_H_
#define SBUF_I_H_

#include "sbuf.h"
#include "rend_dat.h"

#include "static_buffer.h"
#include "stack.h"
#include "llist.h"


  //
  // misc buffers
  //
const int poly_edge_stk_size         =  1024*1024;
const int transparent_poly_stk_size         =  1024*1024;
extern  static_buffer<poly_edge_stk_size>  poly_edge_stk;
extern  static_buffer<512*1024>            rend_dat_stk;
extern  static_buffer<transparent_poly_stk_size> transparent_poly_stk;


extern "C"  double  __cdecl fabs(double);


extern std_fill_func * all_sb_plt_funcs[];

#define  __count_block_dims

#ifdef __count_block_dims
extern int block_4_count,block_8_count,block_16_count;
#endif


inline std_fill_func * get_persp_fill_func(const float min_z_inv, const float dz[2], const int has_transparency )
{
 float worst_dz = fabs(dz[0])+fabs(dz[1]);
 if( min_z_inv  < raster::block_dim_8_min_distance*worst_dz )
 {
#ifdef __count_block_dims
  ++block_4_count;
#endif

  return all_sb_plt_funcs[0*2 + has_transparency ]; 
 }
 else  if( min_z_inv  < raster::block_dim_16_min_distance*worst_dz )
 {
#ifdef __count_block_dims
  ++block_8_count;
#endif
  return all_sb_plt_funcs[1*2 + has_transparency ]; 
 }
 else
 {
#ifdef __count_block_dims
  ++block_16_count;
#endif
  // 16x16 is okay
  return all_sb_plt_funcs[2*2 + has_transparency ]; 
 }
}

//
// internal usage header file
//


//
// for statically allocated structures
//

//
// extract components from a span
//
inline u32 get_y( u32 span)  { return span>>20; }
inline u32 get_x0( u32 span) { return span&1023; }
inline u32 get_x1( u32 span) { return (span>>10)&1023; }


namespace raster
{
 // 2d clip extents
  extern int clip_x_min, clip_x_max, 
             clip_y_min, clip_y_max;

  extern int min_y_span,max_y_span;
  extern u16  * frame_buf_starts[max_y_res];
};




//
// arc types
//

struct sb_arc
{
  sb_arc  * next,* prev;  // prev is left
                          // next is right
  int x;
  int dx;  // fixed point

  sb_arc * end_next;

  int type;  // 0 == left, 1 == right, 2 == interior
  
  inline void remove()
  {
   prev->next = next;
   next->prev = prev;
  }
};

extern raster::poly p_bot;
extern raster::poly p_top;

//
// left arc
//
struct sb_arc_l : sb_arc
 {
  raster::poly * rp;
  raster::poly * l_next; // who was next at start
  raster::poly * l_prev; // who was prev at start
  inline void reset()
   {
     l_next =&p_top;
     l_prev =&p_bot;
     type = 0;

   }
 };
//
// right arc
//
struct sb_arc_r  : sb_arc
{
  raster::poly * lp;
  inline void reset() 
  {
     type = 1;
  }
};

// interior arc
struct sb_arc_i  : sb_arc
 {
  raster::poly * lp;
  raster::poly * rp;
  inline void reset() 
   {
     type = 2;
   }
 };

 

//
// arc trans groups
//
struct arc_trans_group
 {
   sb_arc * enders;
   sb_arc * starters;
 };

extern arc_trans_group arc_trans[max_y_res+1];
struct sbt_poly;
extern sbt_poly *      transparent_poly_starts[max_y_res+1];




//
// span block functions
//

const int max_span_blocks  = 8192;
const int max_transparent_span_blocks  = 4096;
const int span_align_size = 64;
const int span_align_mask = span_align_size-1;

struct span_block;
extern span_block* span_block_stk;
extern span_block * sb_stk_ptr;

extern span_block* transparent_span_block_stk;
extern span_block * transparent_sb_stk_ptr;


// span buffer
const int MAX_SPANS  = 64*2;  // need to double this because of all the text on screen  -jh
typedef u32  span_buffer_type[max_y_res][MAX_SPANS];
extern  span_buffer_type span_buffer;




// span blocks are all aligned on 64 byte boundarys
inline int at_block_end(u32 * ptr) { return (((u32)ptr)&63)==0; }

inline span_block * push_block()
 {
   span_block * rv = sb_stk_ptr++;
   assert( sb_stk_ptr < &span_block_stk[max_span_blocks] );
   return rv;
 }

inline void empty_block_stk()
 {
  sb_stk_ptr = &span_block_stk[2];
 }


inline span_block * push_block_tp()
 {
   span_block * rv = transparent_sb_stk_ptr++;
   assert( transparent_sb_stk_ptr < &transparent_span_block_stk[max_transparent_span_blocks] );
   return rv;
 }

inline void empty_block_stk_tp()
 {
  transparent_sb_stk_ptr = &transparent_span_block_stk[0];
 }


inline u32 pack_span_dat(u32 y, u32 x0, u32 x1 )
{
  return (y << 20)|(x1 << 10)|(x0);
}

// no y
inline u32 pack_span_dat(u32 x0, u32 x1 )
{
  assert( x1 > x0 );
  return (x1 << 10)|(x0);
}

inline void raster::poly::push_span(u32 span)
   {
    if(!at_block_end(cur_span) )
     {
       *cur_span++ = span;
       return;
     }
    span_block * nblock = push_block();
    cur_span[-16] = (u32)&nblock->spans[0];
    cur_span = &nblock->spans[1];
    nblock->spans[0] = span;
    return;
  }



  inline void raster::poly::reset(rend_data * _dat)
   {
     dat = _dat;
     dat->span_buf = push_block();
     cur_span = &dat->span_buf->spans[0]; 
     next = 0;
   }


//
//  linked list node for polys
//
struct poly_ll_link
{
 poly_ll_link * next;
 raster::poly * p;
};

                              


inline void push_poly( poly_ll_link * & list, raster::poly * p)
{
 poly_ll_link * tmp = d_alloc( ((poly_ll_link *)0), rend_dat_stk );
 tmp->p = p;
 llist::add_head( list, tmp );
}




//
// max total polys allowed per screen
//






//
// terminal sb arcs
//
extern sb_arc  a_left, a_right;



//
// poly transitions
//
struct poly_trans
{
  raster::poly * new_top;
  int x;
  inline void set(raster::poly * _new_top, int _x ) { new_top = _new_top; x = _x; } 
};
const int max_poly_transes = 256;
extern min_stack<poly_trans,max_poly_transes> poly_transes;




inline void reset_y_extents()
{
 raster::min_y_span = max_y_res;
 raster::max_y_span = 0;
}




// to avoid under/overflow during interpolation
const float linear_shade_max   = 1.0f - linear_shade_eps;



//
//  glorified 2x2 matrix multiply
//  MSVC handles this macro version better than an inline function
#define SB_LB_TO_SCREEN(D0,DD, S0,S1,S2)                   \
 {                                                \
  float ds0 =  S1 - S0 ;                  \
  float ds1 =  S2 - S0 ;                  \
  DD[0] = xf_mat00*ds0 + xf_mat10*ds1;             \
  DD[1] = xf_mat01*ds0 + xf_mat11*ds1;             \
  D0 = S0 - DD[0]*x0 - DD[1]*y0;    \
 }

#define SB_LB_TO_SCREEN_P(D0,DD, S0,S1,S2)         \
 {                                                \
  float s0 =   S0*v0.z_inv;                       \
  float ds0 =  S1*v1.z_inv - s0 ;                  \
  float ds1 =  S2*v2.z_inv - s0 ;                  \
  DD[0] = xf_mat00*ds0 + xf_mat10*ds1;             \
  DD[1] = xf_mat01*ds0 + xf_mat11*ds1;             \
  D0 = s0 - DD[0]*x0 - DD[1]*y0;    \
 }



#define SB_LB_TO_SCREEN_Z() SB_LB_TO_SCREEN(rd->z0,rd->dz,v0[2],v1[2],v2[2])

//
//  same as above, but for linear shades, which must be 
//  scaled to prevent under/overflow
//
#define SB_LB_TO_SCREEN_SH(D0,DD,S0,S1,S2, x0_off )                   \
 {                                                \
  float ds0 =  S1 - S0;                  \
  float ds1 =  S2 - S0;                  \
  DD[0] = (xf_mat00*ds0 + xf_mat10*ds1)*linear_shade_scale;             \
  DD[1] = (xf_mat01*ds0 + xf_mat11*ds1)*linear_shade_scale;             \
  D0    = S0*linear_shade_scale - (DD[0]) * (x0 + x0_off) - (DD[1])*(y0) + linear_shade_eps;    \
 } 

//
// to avoid numerical unstability.
// tris whose area is less than 2e(min_tri_area_exp) are not drawn
const int min_tri_area_exp = -4;


#define STD_P_SETUP()      \
  float x0 = v0[0];                 \
  float y0 = v0[1];                 \
  float dx1 = (v1[0] - x0);         \
  float dy1 = (v1[1] - y0);         \
                                       \
  float dx2 = (v2[0] - x0);         \
  float dy2 = (v2[1] - y0);         \
                                       \
  /*float area = dx1*dy2 - dy1*dx2;*/      \
   float area = dy1*dx2 - dx1*dy2 ;     \
 /* am I too small?   */               \
   if( comp_exp(area,min_tri_area_exp) == 0)  \
     return 0;                         \
                                       \
  float det_  =  -1.0f/area;            \
                                       \
/* cramer inverse */                   \
  float xf_mat00 =  dy2*det_;          \
  float xf_mat01 = -dx2*det_;          \
  float xf_mat10 = -dy1*det_;          \
  float xf_mat11 =  dx1*det_;          \
  poly * nply = __alloc_poly();




#define STD_P_SETUP_TP()      \
  float x0 = v0[0];                 \
  float y0 = v0[1];                 \
  float dx1 = (v1[0] - x0);         \
  float dy1 = (v1[1] - y0);         \
                                       \
  float dx2 = (v2[0] - x0);         \
  float dy2 = (v2[1] - y0);         \
                                       \
  /*float area = dx1*dy2 - dy1*dx2;*/      \
   float area = dy1*dx2 - dx1*dy2 ;     \
 /* am I too small?   */               \
   if( comp_exp(area,min_tri_area_exp) == 0)  \
     return;                         \
                                       \
  float det_  =  -1.0f/area;            \
                                       \
/* cramer inverse */                   \
  float xf_mat00 =  dy2*det_;          \
  float xf_mat01 = -dx2*det_;          \
  float xf_mat10 = -dy1*det_;          \
  float xf_mat11 =  dx1*det_;          \
  sbt_poly * nply = __alloc_ob(sbt_poly,transparent_poly_stk);


#ifndef NDEBUG
inline void assert_valid( raster::poly * ply)
{
 if( ply )
 {
   assert( (((int)ply)&3) == 0 );
   float z = ply->z_at(5.f,5.f);
   assert( z  <= 1e10 );
   assert( (int)ply >= (int)poly_edge_stk.base );
   assert( (int)ply < (int)poly_edge_stk.p );
 }
}
#else
inline void assert_valid( raster::poly * ) {}
#endif


//
// all polys
//
extern poly_ll_link * alt_polys;
extern poly_ll_link * als_polys;
extern poly_ll_link * afs_polys;
extern poly_ll_link * plt_polys;
extern poly_ll_link      * pls_polys;



#define __alloc_ob(T,STK)  (T*)STK.alloc( sizeof(T) )
#define __alloc_poly()   __alloc_ob(poly,poly_edge_stk)
#define __alloc_edge(T)  __alloc_ob(sb_arc_ ## T ,poly_edge_stk)




#define alloc_rd(T) d_alloc(T,rend_dat_stk)







#define  test_min_int( a,  b )    \
{                                 \
 int delta = (b) - (a);               \
 a += (delta & (delta>>31));      \
}                                 \

#define  test_max_int( a,  b )    \
{                                 \
 int delta = (a) - (b);               \
 a -= (delta & (delta>>31));      \
}                                 


//
// data types for transparent primitives
//
struct sbt_arc
{
  int x;   // 16:16 fixed point
  int dx;  // 16:16 fixed point
  int end_y;  // y at which next arc starts
};



//
// note: these must be convex
//
struct sbt_poly : raster::poly
{
  sbt_poly * next; // next active poly, not sorted.

  int end_y;  // when poly is finished

  float sort_z; // this is a problem.  we currently only sort based on a single z value

  std_fill_func * fill_func;
 
  // left and right arcs
 // we re-use variables here
  inline sbt_arc * & get_left_arc()  { return   *(sbt_arc**)&((raster::poly*)this)->next; }
  inline sbt_arc * & get_right_arc()   { return *(sbt_arc**)&((raster::poly*)this)->prev; }
  inline void push_span(u32 span);
  inline void reset(rend_data * _dat);
};



// transparent
extern sbt_poly * tp_polys;


inline  void sbt_poly::push_span(u32 sp)
   {
    if(!at_block_end(cur_span) )
     {
       *cur_span++ = sp;
       return; 
     }
    span_block * nblock = push_block_tp();
    cur_span[-16] = (u32)&nblock->spans[0];
    cur_span = &nblock->spans[1];
    nblock->spans[0] = sp;
    return;
  }



  inline void sbt_poly::reset(rend_data * _dat)
   {
     dat = _dat;
     dat->span_buf = push_block_tp();
     cur_span = &dat->span_buf->spans[0]; 
     next = 0;
   }



#endif
